package gltest;

import java.awt.Cursor;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;


public class BasicMouseTask3D {
	
	private int operation;
	private static final int NONE = 0, DRAG = 1, ZOOM = 2, ROTATE = 3;
	
	private int startx, starty;
	private int prevx, prevy;
	private double[] startWindow;
	private double requestedAspectRatio;
	
	private Transform3D transform;
	private double virtualSphereRadius;
	
	private Vector3D previousRay, currentRay;
	private long mouseDownTime;
	private boolean startedDragging;
	
	
	private double zoom_startx_real, zoom_starty_real;  // window coords for mouse start postion for zoom operation.
	private double zoom_startx_pixel, zoom_starty_pixel;  // pixel coords for mouse start position for zoom, adjusted to fill window size for sterographic view

	/**
	 * Decides which operation to perform, if any.  For stereographic views, a ZOOM or ROTATE operation is begun
	 * only if the mouse click is actually inside one of the two view rectangles.
	 */
	public boolean doMouseDown(MouseEvent evt, Viewer viewer, Transform3D transform, int width, int height) {
		operation = NONE;
		try {
			startWindow = transform.getWindow();
			this.transform = transform;
			requestedAspectRatio = Math.abs((transform.getYmax_requested() - transform.getYmin_requested()) 
					/ (transform.getXmax_requested() - transform.getXmin_requested()));
		}
		catch (Exception e) {
			return false;
		}
        if (evt.isAltDown() || evt.isShiftDown())
			operation = ZOOM;
		else if (evt.isControlDown() || evt.isMetaDown())
			operation = DRAG;
		else
			operation = ROTATE;
		startx = prevx = evt.getX();
		starty = prevy = evt.getY();

		if (operation == ZOOM) {
			zoom_startx_real = startWindow[0] + startx * ((startWindow[1] - startWindow[0]) / width);
			zoom_starty_real = startWindow[2] + (height-starty) * ((startWindow[3] - startWindow[2]) / height);
			zoom_startx_pixel = startx;
			zoom_starty_pixel = starty;
		}
		else if (operation == ROTATE) {
				virtualSphereRadius = 0.48 * Math.min( Math.abs(transform.getXmax() - transform.getXmin()),
						Math.abs(transform.getYmax() - transform.getYmin()) );
		}
		mouseDownTime = evt.getWhen();
		startedDragging = false;  // won't actualy start until 1/3 second has passed or user moves mouse
		return true;
	}

	/**
	 * Continue an operation that was begun in the <code>doMouseDown</code> method.  The user can change the 
	 * operation in the middle of a mouse drag by pressing and releasing modifier keys.
	 */
	public void doMouseDrag(MouseEvent evt, Viewer viewer, int width, int height) {
		if (operation == NONE)
			return;
		if (!startedDragging && evt.getWhen() - mouseDownTime < 300 
				&& Math.abs(evt.getX() - startx) < 3 && Math.abs(evt.getY() - starty) < 3)
			return;
		startedDragging = true;
		int currentX = evt.getX();
		int currentY = evt.getY();
		switch (operation) {
		case DRAG:
				double offsetX, offsetY;
				double pixelWidth = Math.abs(startWindow[1] - startWindow[0])/width;
				double pixelHeight = Math.abs(startWindow[3] - startWindow[2])/height;
				offsetX =  (startx - currentX) * pixelWidth;
				offsetY = -(starty - currentY) * pixelHeight;
				newWindow(transform,startWindow[0]+offsetX,startWindow[1]+offsetX,
						startWindow[2]+offsetY,startWindow[3]+offsetY);
			break;
		case ZOOM:
			double diff = -(starty - currentY) / 200.0;  
			double mag = Math.exp(diff);  // factor of e every 200 pixels
			double newwidth = (startWindow[1] - startWindow[0]) * mag;
			double newheight = (startWindow[3] - startWindow[2]) * mag;
			double newpixelwidth = newwidth / width;
			double newpixelheight = newheight / height;
			double newxmin = zoom_startx_real - newpixelwidth*zoom_startx_pixel;
			double newxmax = newxmin + newwidth;
			double newymin = zoom_starty_real - newpixelheight*(height-zoom_starty_pixel);
			double newymax = newymin + newheight;
			newWindow(transform,newxmin,newxmax,newymin,newymax);
			break;
		case ROTATE:
			if (prevx == currentX && prevy == currentY)
				break;
			previousRay = mousePointToRay(prevx,prevy);
			currentRay = mousePointToRay(currentX,currentY);
			transform.applyTransvection(previousRay,currentRay);
			prevx = currentX;
			prevy = currentY;
			break;
		}
		viewer.redraw();
	}
	

	public void doMouseUp() {
		if (!startedDragging || operation == NONE)
			return;
		transform = null;
		previousRay = currentRay = null;
		operation = NONE;
	}
	
	
	/**
	 * Converts the mouse point (x,y) to a point on the unit sphere.
	 */
	private Vector3D mousePointToRay(int x, int y) {
		
		Point2D pt = new Point2D.Double(x,y);
		transform.viewportToWindow(pt);

		Vector3D xdir = transform.getImagePlaneXDirection();
		Vector3D ydir = transform.getViewUp();
		
		double vx = pt.getX() * xdir.x + pt.getY() * ydir.x;  // The mouse point as a vector in the image plane.
		double vy = pt.getX() * xdir.y + pt.getY() * ydir.y;
		double vz = pt.getX() * xdir.z + pt.getY() * ydir.z;
		
		double normSquared = vx*vx + vy*vy + vz*vz; 
		
		Vector3D answer;
		
		if (normSquared > virtualSphereRadius*virtualSphereRadius) {  // A point lying in the image plane, on the "equator" of the virtual sphere
			double len = Math.sqrt(normSquared);
			answer = new Vector3D(vx*virtualSphereRadius/len, vy*virtualSphereRadius/len, vz*virtualSphereRadius/len);
		}
		else {  // Add a z-coordinate to (vx,vy,vz) bring the length up to the virtualSphereRadius
			 double z = Math.sqrt(virtualSphereRadius*virtualSphereRadius - normSquared);  
			                        // in view coordinates; vector we want is -z*ViewDirection
			 Vector3D zdir = transform.getViewDirection(); 
			 answer = new Vector3D(vx - z*zdir.x, vy - z*zdir.y, vz -z*zdir.z);
		}
		
		answer.normalize();
		return answer;
		
	}
	
	/**
	 * Used to set a new window for the View that has the same requested aspect ratio
	 * as the current one, and has the specified x and y ranges as the actual ranges
	 * on the window.
	 */
	private void newWindow(Transform3D tr,double xmin, double xmax, double ymin, double ymax) {
		double aspect = Math.abs( (ymax - ymin) / (xmax - xmin) );
		if (aspect > requestedAspectRatio) {  // shrink y range
			double shrinkFactor = requestedAspectRatio / aspect;
			double newHeight = (ymax - ymin) * shrinkFactor;
			double middle = (ymax + ymin) / 2;
			ymin = middle - newHeight / 2;
			ymax = middle + newHeight / 2;
		}
		else { // shrink x range
			double shrinkFactor = aspect / requestedAspectRatio;
			double newWidth = (xmax - xmin) * shrinkFactor;
			double middle = (xmax + xmin) / 2;
			xmin = middle - newWidth / 2;
			xmax = middle + newWidth / 2;
		}
		tr.setLimits(xmin,xmax,ymin,ymax);
	}



}

