package gltest;

import java.awt.geom.Point2D;

/**
 * Holds the transform to be applied the object being viewed
 */
public class Transform3D {
	
	private double clipDistance;
	private Vector3D viewPoint;
	private Vector3D viewDirection;
	private Vector3D imagePlaneXDirection, imagePlaneYDirection;
	
	private double xmin_requested, xmax_requested, ymin_requested, ymax_requested; // Values from setLimits()
	private double xmin,xmax,ymin,ymax; // Actual values used, which might be different because of preserving aspect ratio.
	private int width,height; // The "viewport" in terms of pixels in the drawing area.
	
	public Transform3D(double xmin, double xmax, double ymin, double ymax) {
		setLimits(xmin,xmax,ymin,ymax);
	}
	
	public void setLimits(double xmin, double xmax, double ymin, double ymax) {
		if (xmin == this.xmin && ymin == this.ymin &&
				xmax == this.xmax && ymax == this.ymax)
			return;
		this.xmin = xmin_requested = xmin;
		this.xmax = xmax_requested = xmax;
		this.ymin = ymin_requested = ymin;
		this.ymax = ymax_requested = ymax;
		width = height = 0;
	}
	
	public double[] getWindow() {
		return new double[] { xmin, xmax, ymin, ymax };
	}

	public double getClipDistance() {
		return clipDistance;
	}

	public Vector3D getViewPoint() {
		return viewPoint;
	}

	public Vector3D getViewDirection() {
		return viewDirection;
	}

	public void setViewDirection(Vector3D viewDirection) {
		this.viewDirection = viewDirection;
	}

	public Vector3D getViewUp() {
		return imagePlaneYDirection;
	}

	public void setViewUp(Vector3D viewUp) {
		double projection = viewDirection.dot(viewUp);
		imagePlaneYDirection = new Vector3D(viewUp.x - projection*viewDirection.x, viewUp.y - projection*viewDirection.y, viewUp.z - projection*viewDirection.z);
		// imagePlaneYDirection = viewUp - (viewDirection dot viewUp)*viewDirection  -- projection of viewUp onto image plane
		if (imagePlaneYDirection.norm() < 0.00001)  // try (0,0,1) as the default  ( viewDirection.dot(0,0,1) is just viewDirection.z )
			imagePlaneYDirection = new Vector3D(-viewDirection.z*viewDirection.x, -viewDirection.z*viewDirection.y, 1-viewDirection.z*viewDirection.z);
		if (imagePlaneYDirection.norm() < 0.00001)  // try (0,1,0 if both viewUp and (0,0,1) fail (because they are both too close to being multiples of viewDirection)
			imagePlaneYDirection = new Vector3D(-viewDirection.y*viewDirection.x, 1-viewDirection.y*viewDirection.y,  -viewDirection.y*viewDirection.z);
		imagePlaneYDirection.normalize();
		imagePlaneXDirection = viewDirection.cross(imagePlaneYDirection);
	}

	public double getXmax() {
		return xmax;
	}

	public double getXmin() {
		return xmin;
	}

	public double getYmax() {
		return ymax;
	}

	public double getYmin() {
		return ymin;
	}

	public double getXmax_requested() {
		return xmax_requested;
	}

	public double getYmin_requested() {
		return ymin_requested;
	}

	public double getXmin_requested() {
		return xmin_requested;
	}

	public double getYmax_requested() {
		return ymax_requested;
	}

	public void setClipDistance(double clipDistance) {
		this.clipDistance = clipDistance;
	}

	public void setViewPoint(Vector3D viewPoint) {
		this.viewPoint = viewPoint;
		viewDirection = new Vector3D(this.viewPoint);
		viewDirection.normalize();
		viewDirection.negate();
		double focalLength = viewPoint.norm();
		clipDistance = 0.25*focalLength;
		setViewUp(new Vector3D(0,0,1));
	}

	public void setAspectRatio(int width, int height) {
		double viewAspect = (double)width / height;
		
		double aspect = Math.abs((xmax_requested - xmin_requested) / (ymax_requested - ymin_requested));
		if (viewAspect < aspect) {
			double ycenter = (ymin_requested + ymax_requested) / 2;
			double newheight = (ymax_requested - ymin_requested) * (aspect / viewAspect);
			ymin = ycenter - newheight / 2;
			ymax = ycenter + newheight / 2;
			xmin = xmin_requested;
			xmax = xmax_requested;
		}
		else {
			double xcenter = (xmin_requested + xmax_requested)/2;
			double newwidth = (xmax_requested - xmin_requested) * (viewAspect / aspect);
			xmin = xcenter - newwidth / 2;
			xmax = xcenter + newwidth / 2;
			ymin = ymin_requested;
			ymax = ymax_requested;
		}
		this.width = width;
		this.height = height;
	}
	
	public void viewportToWindow(Point2D pt) {
		if (width == 0)
			return;
		double px = xmin + ((pt.getX()) / width) * (xmax - xmin);
		double py = ymax - ((pt.getY()) / height) * (ymax - ymin);
		pt.setLocation(px,py);
	}

	public Vector3D getImagePlaneXDirection() {
		return imagePlaneXDirection;
	}

	public void applyTransvection(Vector3D e1, Vector3D e2) {  // These must be unit vectors!
		Vector3D e = new Vector3D(e1.x+e2.x, e1.y+e2.y, e1.z+e2.z);
		e.normalize();
		Vector3D temp = new Vector3D();
		reflectInAxis(e,viewDirection,temp);
		reflectInAxis(e1,temp,viewDirection);
		reflectInAxis(e,imagePlaneXDirection,temp);
		reflectInAxis(e1,temp,imagePlaneXDirection);
		reflectInAxis(e,imagePlaneYDirection,temp);
		reflectInAxis(e1,temp,imagePlaneYDirection);
		double vn = viewPoint.norm();
		viewPoint.x = -vn*viewDirection.x;
		viewPoint.y = -vn*viewDirection.y;
		viewPoint.z = -vn*viewDirection.z;
	}

	private void reflectInAxis(Vector3D axis, Vector3D source, Vector3D destination) {
		double s = 2 * (axis.x * source.x + axis.y * source.y + axis.z * source.z);
		destination.x = s*axis.x - source.x;
		destination.y = s*axis.y - source.y;
		destination.z = s*axis.z - source.z;
	}

}
