package gltest;

import gltest.surfaces.Astroidale;
import gltest.surfaces.DiniSurface;
import gltest.surfaces.HyperbolicParaboloid;
import gltest.surfaces.Hyperboloid;
import gltest.surfaces.LissajousSurface;
import gltest.surfaces.MonkeySaddle;
import gltest.surfaces.Paraboloid;
import gltest.surfaces.SnailShell;
import gltest.surfaces.SurfaceParametric;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;

/**
 * A viewer for viewing 3d objects in OpenGL
 */
public class Viewer extends JFrame {
	
	private OpenGLRenderer3D renderer;
	private OpenGLRenderer3DwithPbuffer pRenderer;
	private JPanel drawComponent;
	private SurfaceParametric surface;
	private Transform3D transform;
	
	/**
	 * flag as to whether or not the object is drawn in wireframe or as a solid surface
	 */
	private boolean wireframe;
	
	/**
	 * flag as to whether or not the Pbuffer rendering should be used.
	 */
	private boolean usePbuf = false;
	
	/**
	 * flag as to whether the scene needs to be redrawn
	 */
	private boolean needsRedraw;
	
	/**
	 *flag as to whether grids lines should be shown on the solid surface
	 */
	private boolean showGrid = true;
	
	
	
	public static void main(String[] args) {
		Viewer window = new Viewer();
		window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		window.pack();
		window.setVisible(true);
	}
	
	/**
	 * Viewer constructor sets up to use GLJPanel renderer,
	 * adds mouse handler and menus.
	 */
	public Viewer() {
		super("Surface Viewer -- OpenGL Test");
		renderer = new OpenGLRenderer3D();
				
	    drawComponent = renderer.getRenderComponent(this);
		MouseHandler mouseHandler = new MouseHandler();
		drawComponent.addMouseListener(mouseHandler);
		drawComponent.addMouseMotionListener(mouseHandler);
		setContentPane(drawComponent);
		drawComponent.setPreferredSize(new Dimension(750,500));
		setJMenuBar(createMenus());
		showSurface("Monkey Saddle");
	}
	
	public void setNeedsRedraw(boolean r) {
		needsRedraw = r;
	}
	
	/**
	 * Retrieves the surface data and calls the appropriate renderer.
	 * @param width
	 * @param height
	 */
	public void prepare(int width, int height) {
		if (needsRedraw) {
			Grid3D data = surface.getData();
			if(usePbuf){
				pRenderer.startRender(transform, width, height);
				pRenderer.drawSurface(data, wireframe, showGrid);
				pRenderer.endRender();
			}
			else{
				renderer.startRender(transform, width, height);
				renderer.drawSurface(data, wireframe, showGrid);
				renderer.endRender();
			}
		}
		needsRedraw = false;
	}
	
	/**
	 * Creates a surface menu and an options menu.
	 */
	private JMenuBar createMenus() {
		JMenuBar menus = new JMenuBar();
		JMenu surfaceMenu = new JMenu("Surface");
		menus.add(surfaceMenu);
		JMenu optionMenu = new JMenu("Options");
		menus.add(optionMenu);
		surfaceMenu.add(surfaceItem("Monkey Saddle"));
		surfaceMenu.add(surfaceItem("Paraboloid"));
		surfaceMenu.add(surfaceItem("Snail Shell"));
		surfaceMenu.add(surfaceItem("Hyperboloid"));
		surfaceMenu.add(surfaceItem("Hyperbolic Paraboloid"));
		surfaceMenu.add(surfaceItem("Lissajous Surface"));
		surfaceMenu.add(surfaceItem("Dini Surface"));
		surfaceMenu.add(surfaceItem("Astroidale"));
		final JCheckBoxMenuItem wireframeToggle = new JCheckBoxMenuItem("Wireframe");
		optionMenu.add(wireframeToggle);
		wireframeToggle.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				wireframe = wireframeToggle.getState();
				needsRedraw = true;
				drawComponent.repaint();
			}
		});
		final JCheckBoxMenuItem showGridToggle = new JCheckBoxMenuItem("Show Grid on Surfaces", showGrid);
		optionMenu.add(showGridToggle);
		showGridToggle.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				showGrid = showGridToggle.getState();
				needsRedraw = true;
				drawComponent.repaint();
			}
		});
		final JCheckBoxMenuItem pbufferToggle = new JCheckBoxMenuItem("Use GLPbuffer");
		optionMenu.add(pbufferToggle);
		pbufferToggle.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				usePbuf = pbufferToggle.getState();
				if(usePbuf){
					if(pRenderer == null) {
						pRenderer = new OpenGLRenderer3DwithPbuffer();
						drawComponent = pRenderer.getRenderComponent(Viewer.this);
						MouseHandler mouseHandler = new MouseHandler();
						drawComponent.addMouseListener(mouseHandler);
						drawComponent.addMouseMotionListener(mouseHandler);
					}
					else
						drawComponent = pRenderer.getRenderComponent(Viewer.this);
				}
				else
					drawComponent = renderer.getRenderComponent(Viewer.this);
				setContentPane(drawComponent);
				needsRedraw = true;
				validate();
				drawComponent.repaint();
			}
		});
		return menus;
	}
	
	private JMenuItem surfaceItem(final String name) {
		JMenuItem item = new JMenuItem(name);
		item.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent evt) {
		    	showSurface(name);
		    }
		});
		return item;
	}
	
	/**
	 * Handles surfaces selected from the surface menu.
	 * @param name of surface
	 */
	private void showSurface(String name) {
		if (name.equals("Monkey Saddle"))
			surface = new MonkeySaddle();
		else if (name.equals("Paraboloid"))
			surface = new Paraboloid();
		else if (name.equals("Snail Shell"))
			surface = new SnailShell();
		else if (name.equals("Hyperboloid"))
			surface = new Hyperboloid();
		else if (name.equals("Hyperbolic Paraboloid"))
			surface = new HyperbolicParaboloid();
		else if (name.equals("Lissajous Surface"))
			surface = new LissajousSurface();
		else if (name.equals("Dini Surface"))
			surface = new DiniSurface();
		else if (name.equals("Astroidale"))
			surface = new Astroidale();
		double[] window = surface.getDefaultWindow();
		transform = new Transform3D(window[0], window[1], window[2], window[3]);
		transform.setViewPoint(surface.getDefaultViewPoint());
		needsRedraw = true;
		drawComponent.repaint();
	}
	
	
	public void redraw() {
		drawComponent.repaint();
	}
	
	/**
	 * When the left mouse is pressed and dragged then the object is rotated
	 * When the right mouse is pressed and dragged then the objects position can be changed
	 * When the middle mouse is pressed and dragged then the object is zoom in or out based on mouse drag direction.
	 */
	private class MouseHandler implements MouseListener, MouseMotionListener {
		
		private boolean dragging;
		private BasicMouseTask3D mouseTask = new BasicMouseTask3D();

		public void mouseClicked(MouseEvent arg0) {
		}

		public void mouseEntered(MouseEvent arg0) {
		}

		public void mouseExited(MouseEvent arg0) {
		}

		public void mousePressed(MouseEvent arg0) {
			if (dragging)
				return;
			dragging = mouseTask.doMouseDown(arg0, Viewer.this, transform, drawComponent.getWidth(), drawComponent.getHeight());
		}

		public void mouseReleased(MouseEvent arg0) {
			if (dragging)
				mouseTask.doMouseUp();
			dragging = false;
		}

		public void mouseDragged(MouseEvent arg0) {
			if (dragging)
				mouseTask.doMouseDrag(arg0, Viewer.this, drawComponent.getWidth(), drawComponent.getHeight());
		}

		public void mouseMoved(MouseEvent arg0) {
		}
		
	}

}
