package gltest.surfaces;


import gltest.Grid3D;
import gltest.Vector3D;

/**
 * Defines a suface by a set of parametric equations defined on a rectangular domain
 * in the uv-plane.
 */
abstract public class SurfaceParametric {
	
	protected double umin = -5, umax = 5, vmin = -5, vmax = 5;
	protected int uRes = 15, vRes = 15;
    private Grid3D data;
	private double[] defaultWindow;
	private Vector3D defaultViewPoint;


	abstract protected void setDefaults();
	abstract protected Vector3D surfacePoint(double u, double v);
	

	private final static double epsilon = 0.00005; // for use in derivative functions only
	
	public Grid3D getData() {
		return data;
	}
	
	public SurfaceParametric() {
		setDefaults();
		createData();
	}
	
	/**
	 * Returns an approximation for the partial derivative in the direction u at a specified (u,v) point.
	 */
	public  Vector3D deriv_u (double u, double v) {
		return surfacePoint(u + epsilon, v).minus(surfacePoint(u - epsilon, v)).times(1/(2*epsilon));
	}
	
	/**
	 * Returns an approximation for the partial derivative in the direction v at a specified (u,v) point.
	 */
	public Vector3D deriv_v (double u, double v) {
		return surfacePoint(u, v + epsilon).minus(surfacePoint(u, v - epsilon)).times(1/(2*epsilon));
	}

	/**
	 * Returns the Surface Normal, by default from numerical differentiation.
	 * For Weierstrass surfaces the known Gauss map allows to override the numerical differentiation.
	 */
	public Vector3D surfaceNormal (double u, double v) {
		Vector3D d1 = deriv_u(u,v);
		Vector3D d2 = deriv_v(u,v);
		return d1.cross(d2);
	}
	


	protected void createData() {
		if (data == null)
			data = new Grid3D(uRes*6, vRes*6);  // create a new grid object only if necessary.
		int uCount = data.getUCount();
		int vCount = data.getVCount();
		double u1 = umin;
		double u2 = umax;
		double v1 = vmin;
		double v2 = vmax;
		double du = (u2 - u1) / (uCount - 1);
		double dv = (v2 - v1) / (vCount - 1);
		for (int i = 0; i < uCount; i++) {
			double u = u1 + i*du;
			for (int j = 0; j < vCount; j++) {
				double v = v1 + j*dv;
				data.setVertex(i,j,surfacePoint(u,v));
			}
		}
		for (int i = 0; i <= uCount; i++) {
			double u = u1 + i*du;
			for (int j = 0; j <= vCount; j++) {
				double v = v1 + j*dv;
				Vector3D d1 = deriv_u(u,v);
				Vector3D d2 = deriv_v(u,v);
				Vector3D normal = d1.cross(d2);
				data.setNormal(i,j,normal);
			}
		}
	}
	
	
	protected void setDefaultWindow(double d, double e, double f, double g) {
		defaultWindow = new double[] { d, e, f, g };
	}

	protected void setDefaultViewpoint(Vector3D vector3D) {
		defaultViewPoint = vector3D;
	}

	public Vector3D getDefaultViewPoint() {
		return defaultViewPoint;
	}

	public double[] getDefaultWindow() {
		return defaultWindow;
	}



}
